% Description:
%   Program for testing SVM regression on some relevant data sets.
%
% History:
%   July 12, 2016   : Created by John Reidar Mathiassen.
%   September 15, 2016 : Updated by John Reidar Mathiassen.


% % Modified, consistent.
% load('..\Data\densodatafeaturesmod2consistent.mat');
% X = Xmod;

% Not modified, with pre-computed X,Y,Z.
% load('..\Data\KORREKT.mat');
% load('..\Data\densodatafeatures2.mat');

ods_to_workspace
%remove_Inconsistent


% X-features, y-denso
X = X;
y = y;

y = [y(:,1:12) y(:,15:17)];

yrespnames(1).name = 'X';
yrespnames(2).name = 'Y';
yrespnames(3).name = 'Z';
yrespnames(4).name = 'O1x';
yrespnames(5).name = 'O1y';
yrespnames(6).name = 'O1z';
yrespnames(7).name = 'Ax';
yrespnames(8).name = 'Ay';
yrespnames(9).name = 'Az';
yrespnames(10).name = 'O2x';
yrespnames(11).name = 'O2y';
yrespnames(12).name = 'O2z';
%yrespnames(13).name = 'Figure';
%yrespnames(14).name = 'Preshape';
yrespnames(13).name = 'Significant All 1';
yrespnames(14).name = 'Significant All_2';
yrespnames(15).name = 'Significant All 3';

% Xnew = X;
% Xorig = X;

% X = [X(:,1:3) X(:,7:8)];

% PARTITION TRAINING AND VALIDATION
% set seed to constant, allows for easy comparison
rng(123);

n = size(X,1);
c = cvpartition(n,'HoldOut',0.25); % For truely accurate regression error estimates, this must be done e.g. 10-fold

% Train
Xtr = X(c.training,:);
ytr = y(c.training,:);

% Validation
Xval = X(c.test,:);
yval = y(c.test,:);


% Augment X and y
%%Xaug = X;
%%yaug = y;

Xtraug = Xtr;
ytraug = ytr;
Xvalaug = Xval;
yvalaug = yval;


augstd = 200;
augitnum = 15;

for augit = 1:augitnum
%     Xtmp = X;
%     ytmp = y;
    Xtrtmp = Xtr;
    ytrtmp = ytr;
    Xvaltmp = Xval;
    yvaltmp = yval;
     
%     pert = augstd*randn(size(Xtmp(:,1:3)));
    pertTr = augstd*randn(size(Xtrtmp(:,1:3)));
    pertVal = augstd*randn(size(Xvaltmp(:,1:3)));

%     %    Xaug = [Xaug;X + randn(size(X))];
%     Xtmp(:,1:3) = Xtmp(:,1:3) + pert;
%     ytmp(:,1:3) = ytmp(:,1:3) + pert;
    Xtrtmp(:,1:3) = Xtrtmp(:,1:3) + pertTr;
    ytrtmp(:,1:3) = ytrtmp(:,1:3) + pertTr;
    Xvaltmp(:,1:3) = Xvaltmp(:,1:3) + pertVal;
    yvaltmp(:,1:3) = yvaltmp(:,1:3) + pertVal;
    
     
%     % Append augmentation to temporary
%     Xaug = [Xaug;Xtmp];
%     yaug = [yaug;ytmp];
    Xtraug = [Xtraug;Xtrtmp];
    ytraug = [ytraug;ytrtmp];
    Xvalaug = [Xvalaug;Xvaltmp];
    yvalaug = [yvalaug;yvaltmp];

end

%%X = Xaug;
%%y = yaug;
Xtr = Xtraug;
ytr = ytraug;
Xval = Xvalaug;
yval = yvalaug;


% Finished augmenting
tic;

nval = size(Xval,1);

nd = size(y,2);

epsilon = 0.1;
% Set epsilon as an array, relative to the standard deviation of each
% response variable.
epsilonrel = zeros(nd,1);
for d = 1:nd
    epsilonrel(d) = epsilon*abs(std(ytr(:,d)));
end

cexp = -4:2:8;
gexp = -4:2:8;

k = 5;

mdld = fitcvrsvm(Xtr,ytr,k,epsilonrel,cexp,gexp);
yfit = zeros(nval,nd);
for d = 1:nd
    disp(sprintf('Response variable %u of %u',d,nd));
    disp('---------------------------');
    
    yfit(:,d) = predict(mdld(d).svmmodel,Xval);
    disp(mean(double(abs(yfit(:,d)-yval(:,d)) > epsilon).*abs(yfit(:,d)-yval(:,d))));
end

figure(1);
np = size(X,2);

for p = 1:np
    for d = 1:nd
        subplot(np,nd,p+(d-1)*np);
        plot(X(:,p),y(:,d),'k.');grid on;axis equal;
    end
end

% % figure(2);
% for d = 1:nd
% %     subplot(1,nd,d);
%     figure;
%     plot(yval(:,d),yfit(:,d),'k.');grid on;
% end


for d = 1:nd
    disp(sprintf('Support vector density is %3.2f%%',sum(mdld(d).svmmodel.IsSupportVector)/size(Xtr,1)*100));
end

showtr = 0;

ypredall = zeros(size(yval));
for dd = 1:size(y,2)
    % dd = 5;
    [b, alpha, gamma, mu, sigma, Xsv] = extractrbfsvmparams(mdld(dd).svmmodel,Xtr);
    
    if showtr
        
        [ypred] = predictfromrbfsvmparams(Xmod, b, alpha, gamma, mu, sigma, Xsv);
        [ypredmat] = predict(mdld(dd).svmmodel,Xmod);
        nm = size(Xmod,1);
        
        figure;
        
        subplot(1,2,1);
        title(yrespnames(dd).name);
        hold on;plot(ypred,'k-');plot(y(1:nm,dd),'r--');
        legend('Pred','Truth');
        % plot(ypredmat); hold on;plot(ypred,'k:');plot(yval(:,dd),'r--');
        subplot(1,2,2);
        plot(y(1:nm,dd),ypred,'k.');
        % plot(yval(:,dd),ypred,'k.');
    else
        [ypred] = predictfromrbfsvmparams(Xval, b, alpha, gamma, mu, sigma, Xsv);
        [ypredmat] = predict(mdld(dd).svmmodel,Xval);
        
        ypredall(:,dd) = ypredmat;
        
        figure;
        
        subplot(1,2,1);
        title(yrespnames(dd).name);
        hold on;plot(ypred,'k-');plot(yval(:,dd),'r--');
        legend('Pred','Truth');
        % plot(ypredmat); hold on;plot(ypred,'k:');plot(yval(:,dd),'r--');
        subplot(1,2,2);
        plot(yval(:,dd),ypred,'k.');
        % plot(yval(:,dd),ypred,'k.');
        
        
        
    end
end
% Package rbf SVM parameters into a 2D array for export to LabVIEW

k = size(Xsv,1);
d = size(Xsv,2);

RSVMPARAMS = zeros(k+2,d+1);
RSVMPARAMS(1,1) = b;
RSVMPARAMS(2,1) = gamma;
RSVMPARAMS(1,2:end) = mu;
RSVMPARAMS(2,2:end) = sigma;
RSVMPARAMS(3:end,1) = alpha;
RSVMPARAMS(3:end,2:end) = Xsv;

RSVMPARAMS = [d k zeros(1,d-1);RSVMPARAMS];

writersvmmodel('RSVMPARAMS_TESTING.txt',mdld,Xtr);


% 
% 
% % Test q = R*p + t
% 
% R = [-0.623092738412665       0.718839194676310        0.308262309623673;...
%       0.561122842919132        0.685397282804291        -0.464081587524054;...
%     0.544882184046934        0.116192843676406        0.830423162361596];
% 
% t = [262.374700143223;-4.94598597326984;-539.660015256509];
% 
% Xnew = Xorig;
% 
% for i = 1:size(X,1)
%     p = Xnew(i,1:3)';
%     q = R*p + t;
%     Xnew(i,1:3) = q';
% end
% 
% 
% ynew = zeros(size(y));
% for dd = 1:size(y,2)
%     [b, alpha, gamma, mu, sigma, Xsv] = extractrbfsvmparams(mdld(dd).svmmodel,Xtr);
%     [ypred] = predictfromrbfsvmparams(Xnew, b, alpha, gamma, mu, sigma, Xsv);
% %     [ypredmat] = predict(mdld(dd).svmmodel,Xnew);
%     ynew(:,dd) = ypred;
% end



% Code for plotting result figures.

% Position,
% One orientation
% Gripper


%pos_ori_pres_figure
% figure;
% subplot(3,3,1);
% plot(yval(:,1),ypredall(:,1),'k.');
% axis equal;
% title('p_x');
% xlabel('Demonstrated (mm)');
% ylabel('Estimated (mm)');
% grid on;
% subplot(3,3,2);
% plot(yval(:,2),ypredall(:,2),'k.');
% axis equal;
% title('p_y');
% xlabel('Demonstrated (mm)');
% ylabel('Estimated (mm)');
% grid on;
% subplot(3,3,3);
% plot(yval(:,3),ypredall(:,3),'k.');
% axis equal;
% title('p_z');
% xlabel('Demonstrated (mm)');
% ylabel('Estimated (mm)');
% grid on;
% 
% subplot(3,3,4);
% plot(yval(:,4),ypredall(:,4),'k.');
% axis equal;
% title('d^1_x');
% xlabel('Demonstrated (A.U.)');
% ylabel('Estimated (A.U.)');
% grid on;
% subplot(3,3,5);
% plot(yval(:,5),ypredall(:,5),'k.');
% axis equal;
% title('d^1_y');
% xlabel('Demonstrated (A.U.)');
% ylabel('Estimated (A.U.)');
% grid on;
% subplot(3,3,6);
% plot(yval(:,6),ypredall(:,6),'k.');
% axis equal;
% title('d^1_z');
% xlabel('Demonstrated (A.U.)');
% ylabel('Estimated (A.U.)');
% grid on;
% 
% 
% 
% subplot(3,3,7);
% plot(yval(:,15),ypredall(:,15),'k.');
% axis equal;
% title('spt_1');
% xlabel('Demonstrated (A.U.)');
% ylabel('Estimated (A.U.)');
% grid on;
% subplot(3,3,8);
% plot(yval(:,16),ypredall(:,16),'k.');
% axis equal;
% title('spt_2');
% xlabel('Demonstrated (A.U.)');
% ylabel('Estimated (A.U.)');
% grid on;
% subplot(3,3,9);
% plot(yval(:,17),ypredall(:,17),'k.');
% axis equal;
% title('spt_3');
% xlabel('Demonstrated (A.U.)');
% ylabel('Estimated (A.U.)');
% grid on;


% Save data to a folder, for future reference
newfolder = ['../Run_zero_removed_split_before_' sprintf('%0.0f',augitnum) 'Aug_' sprintf('%0.0f',augstd) 'Std_' datestr(datetime,'yymmdd_HHMM')];
mkdir(newfolder);
save(strcat(newfolder,'/workspace.mat'));
writersvmmodel(strcat(newfolder, '/RSVMPARAMS.txt'),mdld,Xtr);

fignames = {'Features','X','Y','Z','O1x','O1y','O1z','Ax','Ay','Az','O2x','O2y','O2z','SignAll1','SignAll2','SigAll3','pos_ori_pressure'};

h=get(0,'children');
for i=1:length(h)
    savefig(h(i), strcat(newfolder,'/figure_', char(fignames(18-i))));
end






