import numpy as np
import os
import time
import tensorflow as tf
from scipy.special import expit
import matplotlib.pyplot as plt

from CNN.data_loader import CNNData, denorm, cos_sin2deg
from CNN.relativeData import nparray2str

EPOCHS = 15
DROPOUT_KEEP_PROB = 0.5
LEARNING_RATE = 0.0005

OUTPUT_NODES = 9
CHANNELS = 4
PIXELS = 100

LOAD_MODEL = False
TENSORBOARD_LOG = "D:/IProcess_Data/cnn_logs/"
TENSORBOARD_MODELS = "D:/IProcess_Data/cnn_models/"
save_file = "model"

if LOAD_MODEL:
    model_file = "model_1497258021"
else:
    model_file = save_file + "_" + str(round(time.time()))


# TODO: Build network

def save_model(save,file_name, step):
    # must end in .ckpt"
    model_file_name = TENSORBOARD_MODELS + file_name + ".ckpt"
    save.save(sess, model_file_name, global_step=step)
    print("*" * 100)
    print("Model saved in file: %s" % model_file_name)
    print("*" * 100)


def load_model(save,filename):
    model_file_name = TENSORBOARD_MODELS + filename + ".ckpt"
    last = tf.train.latest_checkpoint(TENSORBOARD_MODELS)
    print("#"*100)
    print(last)
    save.restore(sess, model_file_name)

    print("*" * 100)
    print("Model loaded from file: %s" % model_file_name)
    print("*" * 100)


def launch_tensorboard(LOG_DIRECTORY, chrome=False):
    temp = os.getcwd()
    os.chdir("D:\Anaconda\envs\python35")
    # START TENSORBOARD AND CHROME
    os.system("start python -m tensorflow.tensorboard --logdir=" + LOG_DIRECTORY)

    if chrome:
        os.system("start chrome http://192.168.1.201:6006/")
    os.chdir(temp)


def denomralize_outputs(output):
    output[:3] = denorm(output[:3], data.norm_min, data.norm_max)
    if len(output) >= 3:
        output[3:] = denorm(output[3:], -1, 1)
        output = cos_sin2deg(output)
    return output


def show_some_predictions(session, images, labels, number=5):
    predictions = session.run(net, feed_dict={x: images[:number]})
    predictions = expit(predictions)
    for n in range(number):
        if False:
            print("Actual:     \t" + "".join(["%15.7f\t" % i for i in labels[n]]))
            print("Predictions:\t" + "".join(["%15.7f\t" % i for i in predictions[n]]))
        print("Actual:     \t" + "".join(["%15.7f\t" % i for i in denomralize_outputs(labels[n])[0]]))
        print("Predictions:\t" + "".join(["%15.7f\t" % i for i in denomralize_outputs(predictions[n])[0]]))


    print("\n")


x = tf.placeholder(tf.float32, [None, PIXELS, PIXELS, CHANNELS], name="X")
y = tf.placeholder(tf.float32, [None, OUTPUT_NODES], name="Y_")
keep_prob = tf.placeholder("float")

'''
INPUT: [batch, height, width, channels]
FILTER: [filter_height, filter_width, in_channels, out_channels]
STRIDES: ints, stride for eac dimension of INPUT
PADDING: "SAME": zero-padding, "VALID": no padding
'''


def batch_normal(net_in, count):
    if False:
        print("bacth normalizing")
        return tf.layers.batch_normalization(inputs=net_in, name="norm" + str(count))
    else:
        return net_in


net = batch_normal(x, 1)
print("input\t\t", x.shape)
net = tf.layers.conv2d(inputs=net, filters=16, kernel_size=[3, 3], strides=[1, 1], padding='same',
                       activation=tf.nn.relu, name="conv0")
print("conv layer\t", net.shape)
net = batch_normal(net, 2)

net = tf.layers.conv2d(inputs=net, filters=32, kernel_size=[3, 3], strides=[1, 1], padding='same',
                       activation=tf.nn.relu, name="conv1")
print("conv layer\t", net.shape)
net = batch_normal(net, 98)


net = tf.layers.max_pooling2d(inputs=net, pool_size=2, strides=2, padding='same', name="poolxx")
print("pooling\t", net.shape)
net = batch_normal(net, 99)


net = tf.layers.conv2d(inputs=net, filters=64, kernel_size=[3, 3], strides=[1, 1], padding='same',
                       activation=tf.nn.relu, name="conv2")
print("conv layer\t", net.shape)

net = tf.layers.max_pooling2d(inputs=net, pool_size=2, strides=2, padding='same', name="pool1")
print("pooling\t", net.shape)
net = batch_normal(net, 3)

net = tf.layers.conv2d(inputs=net, filters=128, kernel_size=[3, 3], strides=[1, 1], padding='same',
                       activation=tf.nn.relu, name="con3")
print("conv layer\t", net.shape)
net = batch_normal(net, 4)
net = tf.layers.max_pooling2d(inputs=net, pool_size=2, strides=2, padding='same', name="pool2")
print("pooling\t", net.shape)
net = batch_normal(net, 5)

net = tf.contrib.layers.flatten(net)
print("flatten\t\t", net.shape)
net = batch_normal(net, 6)

net = tf.layers.dense(inputs=net, units=4096, activation=tf.nn.relu, name="dense1")
print("fully connected\t", net.shape)
net = batch_normal(net, 7)

# net = tf.layers.dropout(inputs=net, rate=DROPOUT_KEEP_PROB, name="dropout")
# print("dropout\t", net.shape)

net = tf.layers.dense(inputs=net, units=OUTPUT_NODES, name="dense2")
print("output\t", net.shape)

loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(logits=net, labels=y), name="loss")
tf.summary.scalar("Loss", loss)
optimizer = tf.train.AdamOptimizer(learning_rate=LEARNING_RATE, name="optimizer").minimize(loss)
# optimizer = tf.train.GradientDescentOptimizer(learning_rate=0.01).minimize(loss)

abs_distance = tf.reduce_mean(tf.sqrt(tf.reduce_sum(tf.square(tf.subtract(net[:,:3], y[:,:3])), axis=1)))
tf.summary.scalar("Absolute Distance", abs_distance)


sess = tf.Session()
saver = tf.train.Saver(max_to_keep=2)
init = tf.global_variables_initializer()

mergesum = tf.summary.merge_all()
writer = tf.summary.FileWriter(TENSORBOARD_LOG + model_file)
#launch_tensorboard(TENSORBOARD_LOG)
writer.add_graph(sess.graph)

# Load data
data = CNNData(PIXELS)

# Train network
sess.run(init)
print("**TRAINING**\nBatch size:  %s\t" % data.batch_size, len(data.index_train)/data.batch_size)

if LOAD_MODEL:
    print("LOADING")
    load_model(saver, model_file)
else:
    print("BEGIN TRAIN")
    for epoch in range(EPOCHS):
        batches = data.get_training_batches()
        labels = batches['label']
        images = batches['img']
        print("Epoch: ",epoch, end='\t')
        for batch in range(len(batches['split'])):
            images[batch] = np.reshape(images[batch], [-1, PIXELS, PIXELS, CHANNELS])

            _, loss_b, summary = sess.run([optimizer, loss, mergesum],
                                          feed_dict={x: images[batch], y: labels[batch][:, :OUTPUT_NODES]})

            print(batch, end='..')
        print('DONE')
        writer.add_summary(summary, epoch)
        print(loss_b)

        if epoch % 5 == 0 or epoch == EPOCHS-1:
            if False:
                plt.imshow(images[0, :, :, :])
                plt.show()

            train_images = images[0]
            train_label = labels[0][:, :OUTPUT_NODES]
            train_loss = sess.run(abs_distance, feed_dict={x: train_images, y: train_label})

            test_data = data.get_testing_batch()
            test_images = np.reshape(test_data['img'], [-1, PIXELS, PIXELS, CHANNELS])
            test_label = test_data['label'][:, :OUTPUT_NODES]
            test_loss = sess.run(abs_distance, feed_dict={x: test_images[:100], y: test_label[:100]})

            #print('Epoch: %3.d Train-loss:%15f Test-loss:%15f' % (epoch, train_loss, test_loss))
            print('Epoch: %3.d Train-Dist:%15f Test-Dist:%15f' % (epoch, train_loss, test_loss))

            show_some_predictions(sess, train_images, train_label)
            show_some_predictions(sess, test_images, test_label)

            # TODO: save ckpt after every 10th epoch
            save_model(saver, model_file, epoch)

    save_model(saver, model_file, epoch)




# Todo: test network
print("**TESTING**\n")

test_data = data.get_testing_batch()
test_images = np.reshape(test_data['img'], [-1, PIXELS, PIXELS, CHANNELS])
test_label = test_data['label'][:, :OUTPUT_NODES]
show_some_predictions(sess, test_images[:100], test_label[:100])


#predictions = sess.run(net, feed_dict={x: test_images})
#predictions = expit(predictions)
#predictions = denomralize_outputs(predictions)

# print(predictions)




# todo:      Convert back to DENSO coordinates

## get output from CNN
##data.relativeData_object.set_relative_6DOF(CNN output)
