import numpy as np
import tensorflow as tf
from scipy.special import expit
import matplotlib.pyplot as plt

from CNN.data_loader import CNNData

EPOCHS = 2000
MIN_LOSS = 100
DROPOUT_KEEP_PROB = 0.5
LEARNING_RATE = 0.00001
OUTPUT_NODES = 3
CHANNELS = 1


# TODO: Build network

def show_some_predictions(session, images, labels, number=5):
    predictions = session.run(output, feed_dict={x: images})
    for i in range(number):
        a = "Actual:     \t"
        p = "Predictions:\t"
        for j in range(OUTPUT_NODES):
            a += "%15.7f \t" %labels[i][j]
            p += "%15.7f*\t" %predictions[i][j]

        print(a)
        print(p)
    print("\n")


x = tf.placeholder(tf.float32, [None, 28, 28, 1], name="X")
y = tf.placeholder(tf.float32, [None, OUTPUT_NODES], name="Y_")
keep_prob = tf.placeholder("float")



'''
INPUT: [batch, height, width, channels]
FILTER: [filter_height, filter_width, in_channels, out_channels]
STRIDES: ints, stride for eac dimension of INPUT
PADDING: "SAME": zero-padding, "VALID": no padding
'''

print(x.shape)
print("conv")
conv = tf.nn.conv2d(input=x, filter=tf.random_normal([3, 3, CHANNELS, 32]), strides=[1, 1, 1, 1], padding='SAME')
print(conv.shape)
conv = tf.nn.relu(conv)
print(conv.shape)

print("conv1")
conv1 = tf.nn.conv2d(input=conv, filter=tf.random_normal([3, 3, 32, 64]), strides=[1, 1, 1, 1], padding='SAME')
print(conv1.shape)
conv1 = tf.nn.relu(conv1)
print(conv1.shape)

print("pool1")
pool1 = tf.nn.max_pool(conv1, ksize=[1, 2, 2, 1], strides=[1, 2, 2, 1], padding='SAME')
print(pool1.shape)

print("conv2")
conv2 = tf.nn.conv2d(input=pool1, filter=tf.random_normal([3, 3, 64, 128]), strides=[1, 1, 1, 1], padding='SAME')
print(conv2.shape)
conv2 = tf.nn.relu(conv2)
print(conv2.shape)

print("pool2")
pool2 = tf.nn.max_pool(conv2, ksize=[1, 2, 2, 1], strides=[1, 2, 2, 1], padding='SAME')
print(pool2.shape)
print("flatten")
flat1 = tf.contrib.layers.flatten(pool2)
print(flat1.shape)

print("fc1")
fc1 = tf.layers.dense(inputs=flat1, units=1024, activation=tf.nn.relu)
print(fc1.shape)

print("dropout")
dropout = tf.nn.dropout(fc1, keep_prob=DROPOUT_KEEP_PROB)
print(dropout.shape)

print("output")
output = tf.layers.dense(inputs=dropout, units=3)

print(output.shape)
#coord_distance = tf.sqrt(tf.reduce_sum(tf.square(tf.subtract(output, y)), axis=1))

loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(logits=output, labels=y))
optimizer = tf.train.AdamOptimizer(learning_rate=LEARNING_RATE).minimize(loss)
#optimizer = tf.train.GradientDescentOptimizer(learning_rate=0.001).minimize(loss)


init = tf.global_variables_initializer()

# Load data
data = CNNData()


# Train network
sess = tf.Session()
sess.run(init)
batches = data.get_training_batches()['img']
print("**TRAINING**\nBatch size:  %s\t\t\t(BatchCount, LastBatch)" %data.batch_size, divmod(len(data.index_train),data.batch_size))

for epoch in range(EPOCHS):
    batches = data.get_training_batches()
    for b in range(len(batches['split'])):
        label = batches['label'][b]
        img = batches['img'][b]
        img = np.reshape(img, [-1, 28, 28, CHANNELS])

        _, loss_b = sess.run([optimizer, loss], feed_dict={x: img, y: label[:, :OUTPUT_NODES]})

    if epoch % (25) == 0:

        train_data = data.get_all_training()
        train_images = np.reshape(train_data['img'], [-1, 28, 28, CHANNELS])

        plt.imshow(train_images[0,:,:,0], cmap="gray")
        plt.show()

        train_label = train_data['label'][:, :OUTPUT_NODES]
        train_loss = sess.run(loss, feed_dict={x: train_images, y: train_label})

        test_data = data.get_testing_batch()
        test_images = np.reshape(test_data['img'], [-1, 28, 28, CHANNELS])
        test_label = test_data['label']
        test_loss = sess.run(loss, feed_dict={x: test_images, y: test_label[:, :OUTPUT_NODES]})

        print('Epoch: %3.d Train-loss:%10.d Test-loss:%10.d' % (epoch, train_loss, test_loss))

        show_some_predictions(sess, train_images, train_label)
        #show_some_predictions(sess, test_images, test_label)

        #if train_loss <= MIN_LOSS:
         #   break

# Todo: test network
print("**TESTING**\n")

test_data = data.get_testing_batch()
test_images = np.reshape(test_data['img'], [-1, 28, 28, CHANNELS])
test_label = test_data['label']
test_loss = sess.run(loss, feed_dict={x: test_images, y: test_label[:, :OUTPUT_NODES]})

print('Epoch: %3.d Train-loss:%10.d Test-loss:%10.d' % (epoch, train_loss, test_loss))


# testing = data.get_testing_batch()
# test_loss, pred = sess.run([loss, output], feed_dict={x: np.reshape(testing['img'], [-1, 28, 28, 1]),y: testing['label'][:, :3]})
# print(test_loss)

# print(nparray2str(pred))





# todo:      Convert back to DENSO coordinates

## get output from CNN
##data.relativeData_object.set_relative_6DOF(CNN output)
