import numpy as np
import os
import time
import tensorflow as tf
from scipy.special import expit
import matplotlib.pyplot as plt

from CNN.data_loader import CNNData, denorm, cos_sin2deg, regression, save_prediction_to_file
from CNN.relativeData import nparray2str

EPOCHS = 10
LEARNING_RATE = 0.001

OUTPUT_NODES = 9
CHANNELS = 4
PIXELS = 128

NORMALISE_BETWEEN_LAYERS = True
AUG_90 = False
AUG_180 = False

LOAD_MODEL = False
TENSORBOARD_LOG = "D:/IProcess_Data/cnn_logs/"
TENSORBOARD_MODELS = "D:/IProcess_Data/cnn_models/"
save_file = "model"

if LOAD_MODEL:
    # REMEMBER TO USE WHOLE FILE, ex: model_1497258021.ckpt-14
    model_file = "model_1497508826.ckpt-15"
    #model_file = "model_1497508826.ckpt-15"
    # no need to augment when loadin model as only unagmented values are used
    AUG_90 = AUG_180 = False
else:
    model_file = save_file + "_" + str(round(time.time()))


# TODO: Build network

def save_model(sess, saver, file_name, step):
    # must end in .ckpt"
    model_file_name = TENSORBOARD_MODELS + file_name + ".ckpt"
    saver.save(sess, model_file_name, global_step=step)
    print("*" * 100)
    print("Model saved in file: %s" % model_file_name)
    print("*" * 100)


def load_model(sess, filename):
    model_file_name = TENSORBOARD_MODELS + filename
    saver = tf.train.import_meta_graph(model_file_name + ".meta")
    saver.restore(sess, model_file_name)

    print("*" * 100)
    print("Model loaded from file: %s" % model_file_name)
    print("*" * 100)


def launch_tensorboard(LOG_DIRECTORY, chrome=False):
    temp = os.getcwd()
    os.chdir("D:\Anaconda\envs\python35")
    # START TENSORBOARD AND CHROME
    os.system("start python -m tensorflow.tensorboard --logdir=" + LOG_DIRECTORY)

    if chrome:
        os.system("start chrome http://192.168.1.201:6006/")
    os.chdir(temp)


def denomralize_outputs(output):
    output_copy = np.copy(output)
    output_copy[:, :3] = denorm(output_copy[:, :3], data.norm_min, data.norm_max)
    if output_copy.shape[1] > 3:
        output_copy[:, 3:] = denorm(output_copy[:, 3:], -1, 1)
        output_copy = cos_sin2deg(output_copy)
    return output_copy


def show_some_predictions(session, images, labels, number=5):
    predictions = session.run(net, feed_dict={x: images[:number]})

    predictions = expit(predictions)
    conv_label = denomralize_outputs(labels)
    conv_pred = denomralize_outputs(predictions)

    for n in range(number):
        print("Actual:     \t" + "".join(["%15.7f\t" % i for i in conv_label[n]]))
        print("Predictions:\t" + "".join(["%15.7f\t" % i for i in conv_pred[n]]))
        print("Distance:   \t\t" + str(np.sqrt(np.sum(np.square(np.subtract(conv_label[n, :3], conv_pred[n, :3]))))),
              end="\t")
        print("Rotation    \t" + str(np.subtract(np.mod(conv_pred[n, 5], 180), np.mod(conv_label[n, 5], 180))))


    print("\n")


# Load data
data = CNNData(pixels=PIXELS, aug90=AUG_90, aug180=AUG_180)

# Start TensorFlow session
sess = tf.Session()
#launch_tensorboard(TENSORBOARD_LOG)



"""
Generating a model or using a previously trained model
1. The shape and size of the model
2a. Training the network on the training data
2b. Validating the trained model on test data that model has not seen before
2c. Saving model so that it is possible to resume if interupted or for comparison
2d. Save final model when finished
3. If a previously trained model is chosen, loads this model, and creates various variables needed to function
4. Tests final model, this is either a loaded model, or the model from final epoch


"""
if not LOAD_MODEL:
    """     1.      """
    x = tf.placeholder(tf.float32, [None, PIXELS, PIXELS, CHANNELS], name="X")
    tf.add_to_collection("X", x)
    y = tf.placeholder(tf.float32, [None, OUTPUT_NODES], name="Y_")
    tf.add_to_collection("Y_", y)


    def batch_normal(net_in, count):
        if NORMALISE_BETWEEN_LAYERS:
            print("bacth normalizing")
            return tf.layers.batch_normalization(inputs=net_in, name="norm" + str(count))
        else:
            return net_in


    net = batch_normal(x, 0)
    print("input\t\t", x.shape)

    net = tf.layers.conv2d(inputs=net, filters=32, kernel_size=[3, 3], strides=[1, 1], padding='same',
                           activation=tf.nn.relu, name="conv0")
    print("conv layer\t", net.shape)
    net = batch_normal(net, 1)

    net = tf.layers.conv2d(inputs=net, filters=32, kernel_size=[3, 3], strides=[1, 1], padding='same',
                           activation=tf.nn.relu, name="conv1")
    print("conv layer\t", net.shape)
    net = batch_normal(net, 2)

    net = tf.layers.max_pooling2d(inputs=net, pool_size=2, strides=2, padding='same', name="pool1")
    print("pooling\t", net.shape)
    net = batch_normal(net, 3)

    net = tf.layers.conv2d(inputs=net, filters=64, kernel_size=[3, 3], strides=[1, 1], padding='same',
                           activation=tf.nn.relu, name="conv2")
    print("conv layer\t", net.shape)

    net = tf.layers.max_pooling2d(inputs=net, pool_size=2, strides=2, padding='same', name="pool2")
    print("pooling\t", net.shape)
    net = batch_normal(net, 4)

    net = tf.layers.conv2d(inputs=net, filters=128, kernel_size=[3, 3], strides=[1, 1], padding='same',
                           activation=tf.nn.relu, name="conv3")
    print("conv layer\t", net.shape)
    net = batch_normal(net, 5)
    net = tf.layers.max_pooling2d(inputs=net, pool_size=2, strides=2, padding='same', name="pool3")
    print("pooling\t", net.shape)
    net = batch_normal(net, 6)

    net = tf.contrib.layers.flatten(net)
    print("flatten\t\t", net.shape)
    net = batch_normal(net, 7)

    net = tf.layers.dense(inputs=net, units=4096, activation=tf.nn.relu, name="dense1")
    print("fully connected\t", net.shape)
    net = batch_normal(net, 8)

    # net = tf.layers.dropout(inputs=net, rate=DROPOUT_KEEP_PROB, name="dropout")
    # print("dropout\t", net.shape)

    net = tf.layers.dense(inputs=net, units=OUTPUT_NODES, name="dense2")
    print("output\t", net.shape)

    tf.add_to_collection("net", net)

    loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(logits=net, labels=y), name="loss")
    tf.add_to_collection("loss", loss)
    tf.summary.scalar("Loss", loss)
    optimizer = tf.train.AdamOptimizer(learning_rate=LEARNING_RATE, name="optimizer").minimize(loss)
    # optimizer = tf.train.GradientDescentOptimizer(learning_rate=0.01).minimize(loss)

    abs_distance = tf.reduce_mean(tf.sqrt(tf.reduce_sum(tf.square(
        tf.subtract(denorm(tf.sigmoid(net[:, :3]), data.norm_min, data.norm_max),
                    denorm(y[:, :3], data.norm_min, data.norm_max))), axis=1)))
    # abs_distance = tf.reduce_mean(tf.sqrt(tf.reduce_sum(tf.square(tf.subtract(net[:, :3], y[:, :3])), axis=1)))
    tf.summary.scalar("Absolute Distance", abs_distance)
    tf.add_to_collection("abs_distance", abs_distance)

    mergesum = tf.summary.merge_all()
    writer = tf.summary.FileWriter(TENSORBOARD_LOG + model_file)
    writer.add_graph(sess.graph)

    init = tf.global_variables_initializer()
    sess.run(init)
    model_saver = tf.train.Saver(max_to_keep=3)

    """     2a.      """
    # Train network
    print("**TRAINING**\nBatch size:  %s\t" % data.batch_size, len(data.index_train) / data.batch_size)
    print("BEGIN TRAIN")
    counter = 0
    for epoch in range(EPOCHS):
        batches = data.get_training_batches()
        labels = batches['label']
        images = batches['img']
        print("Epoch: ", epoch, end='\t')
        for batch in range(len(batches['split'])):
            images[batch] = np.reshape(images[batch], [-1, PIXELS, PIXELS, CHANNELS])

            _, loss_b, summary = sess.run([optimizer, loss, mergesum],
                                          feed_dict={x: images[batch], y: labels[batch][:, :OUTPUT_NODES]})
            writer.add_summary(summary, counter)
            counter += 1
            print(loss_b, end='  ')
        print('DONE')
        print(loss_b)

        """     2b.     """
        if epoch % 5 == 0 or epoch == EPOCHS - 1:
            if False:
                plt.imshow(images[0, :, :, :])
                plt.show()

            train_images = images[0]
            train_label = labels[0][:, :OUTPUT_NODES]
            train_loss = sess.run(abs_distance, feed_dict={x: train_images, y: train_label})

            test_data = data.get_testing_batch()
            test_images = np.reshape(test_data['img'], [-1, PIXELS, PIXELS, CHANNELS])
            test_label = test_data['label'][:, :OUTPUT_NODES]
            test_loss = sess.run(abs_distance, feed_dict={x: test_images[:50], y: test_label[:50]})

            # print('Epoch: %3.d Train-loss:%15f Test-loss:%15f' % (epoch, train_loss, test_loss))
            print('Epoch: %3.d (%4.d) Train-Dist:%15f Test-Dist:%15f' % (epoch, counter, train_loss, test_loss))

            show_some_predictions(sess, train_images, train_label)
            show_some_predictions(sess, test_images, test_label)

            """     2c.     """
            if epoch != EPOCHS - 1:
                save_model(sess, model_saver, model_file, epoch)
    """     2d.     """
    save_model(sess, model_saver, model_file, EPOCHS)

else:
    """     3.      """
    print("LOADING")

    load_model(sess, model_file)

    for i in tf.trainable_variables(): print(i)
    #    for i in tf.get_default_graph().get_all_collection_keys(): print(i)

    net = tf.get_collection("net")[0]
    x = tf.get_collection("X")[0]
    y = tf.get_collection("Y_")[0]
    abs_distance = tf.get_collection("abs_distance")[0]

"""     4.      """
print("**TESTING**\n")

test_data = data.get_testing_batch()
test_images = np.reshape(test_data['img'], [-1, PIXELS, PIXELS, CHANNELS])
test_label = test_data['label'][:, :OUTPUT_NODES]
show_some_predictions(sess, test_images, test_label)

"""
Generating complete predictions as well as regression and csv files

1. calculate complete dataset
2. calculate predicted denso data by setting relative data
3a. Generate regression tables
3b. Save regression of validation/test to file for review
3c. Generate csv_data for visualization and denso control
"""

"""     1.      """
complete_images = data.png_array[:data.data_count]
complete_prediction = np.empty((0, 9))
b_size = 50
for i in range(int(data.data_count / b_size) + 1):
    temp = complete_images[i * b_size:(i + 1) * b_size]
    temp_pred = sess.run(net, feed_dict={x: temp})

    complete_prediction = np.vstack((complete_prediction, temp_pred))
complete_prediction = expit(complete_prediction)
complete_prediction = denomralize_outputs(complete_prediction)

complete_labels = data.relative_6dof[:data.data_count]
complete_labels = denomralize_outputs(complete_labels)

"""     2.      """
data.relativeData_object.set_relative_6dof(complete_prediction)
orig_denso = data.relativeData_object.get_denso_6dof()
pred_denso = data.relativeData_object.get_predicted_denso_6dof()

"""     3a.     """
print("\nRelative regression results:")
regression(complete_labels[data.indexes], complete_prediction[data.indexes], "All")
regression(complete_labels[data.index_train_original], complete_prediction[data.index_train_original], "Train")
reg_relative_validation = regression(complete_labels[data.index_test_original], complete_prediction[data.index_test_original], "Test")

print("\nDenso regression results:")
regression(orig_denso[data.indexes], pred_denso[data.indexes], "All")
regression(orig_denso[data.index_train_original], pred_denso[data.index_train_original], "Train")
reg_denso_validation = regression(orig_denso[data.index_test_original], pred_denso[data.index_test_original], "Test")



"""     3b.      """
with open("regression_overview.csv", "a") as f:
    string = model_file
    for r in reg_denso_validation:
        string += "\t" + str(r)
    f.write(string + "\n")

with open("regression_overview_relative.csv", "a") as f:
    string = model_file
    for r in reg_relative_validation:
        string += "\t" + str(r)
    f.write(string + "\n")


"""     3c.      """
print("\nSaving Predictions to file")
save_prediction_to_file(pred_denso)
save_prediction_to_file(pred_denso, filepath="cnn_csv_files/" + model_file +".csv")





# todo:      Convert back to DENSO coordinates





## get output from CNN
##data.relativeData_object.set_relative_6DOF(CNN output)
