import numpy as np


class CSVData:
    def __init__(self, filename, densodata=True, example_count=None):
        self.filename = filename
        self.densodata = densodata
        self.examples = None
        self.labels = None
        self.values = None

        self.coordinates = None
        self.magnification = None
        self.rotation = None

        self.open_file(filename, example_count)

        self.extract_coordinates()
        self.extract_magnification()
        self.extract_rotation()

    def __str__(self):
        return 'CSVData@' + self.filename

    def open_file(self, filename, limit):
        labels, values = csv_to_nparray(filename=filename)

        self.labels = labels[:limit]
        self.values = values[:limit]
        self.examples = self.values.shape[0]

    def extract_coordinates(self):
        if self.densodata:
            self.coordinates = self.values[:, :3]
        else:
            self.coordinates = self.values[:, 4:7]

    def extract_magnification(self):
        if not self.densodata:
            self.magnification = self.values[:, 3]
            self.magnification = np.reshape(self.magnification, (self.magnification.shape[0], 1))

    def extract_rotation(self):
        if self.densodata:
            self.rotation = self.values[:, 3:6]


def csv_to_nparray(filename):
    """
    converts the .csv data to a numpy.nparray
    returns both Labels as L, and Values as V
    """
    file = open(filename, "r")
    data = [x.split("\t") for x in file]
    data_transformed = np.asarray(data).T

    labels = data_transformed[0, :]
    values = data_transformed[1:, :].astype(float)

    return labels, values
