import numpy as np

'''
coord_object:   center of mass of object
coord_gripper:  coordinates of gripper
magnification:  how much the object was zoomed to fit 227*227
rotation:       rotaion of gripper

coord_relative: gripper location in relation to object
distance:       distance(XYZ) between object and gripper coordinates

'''

use_multiply = True


class RelativeData:
    def __str__(self):
        return 'relative@' + str(self.__hash__())

    def __init__(self, object_coordinates, object_magnification):
        self.coord_object = object_coordinates
        self.magnification = np.reshape(object_magnification, (object_magnification.shape[0], 1))

        self.coord_relative = None
        self.coord_gripper = None
        self.rotation = None

        self.prediction_coord_relative = None
        self.prediction_coord_gripper = None
        self.prediction_rotation = None

        self.indexes_under_limit = None

    def set_gripperdata(self, gripper_coordinates, gripper_rotation, limit=100):
        self.coord_gripper = gripper_coordinates
        self.rotation = gripper_rotation
        self.set_indexes_under_limit(limit)
        self.min_max()

    def min_max(self):
        relative = self.get_relative_6dof()[self.indexes_under_limit]
        print("Min and Max of relative data")
        print("min\t",nparray2str([np.min(relative, axis=0)]))
        print("max\t",nparray2str([np.max(relative, axis=0)]))

    def calculate_coord_relative(self):
        self.coord_relative = np.subtract(self.coord_gripper, self.coord_object)
        if use_multiply:
            self.coord_relative = np.multiply(self.coord_relative, self.magnification)

    def calculate_prediction_coord_denso_gripper(self):
        # if use_multiply:
        #     self.coord_relative = np.divide(self.coord_relative, self.magnification)
        # self.coord_gripper = np.add(self.coord_object, self.coord_relative)
        if use_multiply:
            self.prediction_coord_relative = np.divide(self.prediction_coord_relative, self.magnification)
        self.prediction_coord_gripper = np.add(self.coord_object, self.prediction_coord_relative)

    def set_relative_6dof(self, cnn_output):
        # self.coord_relative = cnn_output[:, :3]
        # self.rotation = cnn_output[:, 3:6]
        # self.calculate_coord_denso_gripper()
        self.prediction_coord_relative = cnn_output[:, :3]
        self.prediction_rotation = cnn_output[:, 3:6]
        self.calculate_prediction_coord_denso_gripper()

    def get_relative_6dof(self):
        return np.hstack((self.coord_relative, self.rotation))

    def get_predicted_denso_6dof(self):
        return np.hstack((self.prediction_coord_gripper, self.prediction_rotation))

    def get_denso_6dof(self):
        return np.hstack((self.coord_gripper, self.rotation))

    def calculate_relative_distance(self):
        self.calculate_coord_relative()
        if use_multiply:
            distance = np.square(np.divide(self.coord_relative, self.magnification))
        else:
            distance = np.square(self.coord_relative)
        distance = np.sum(distance, axis=1)
        return np.sqrt(distance).reshape(distance.shape[0], 1)

    def index_of_non_extremes(self, max_distance):
        distances = self.calculate_relative_distance()
        non_extremes = []
        for i in range(len(distances)):
            if distances[i] < max_distance:
                non_extremes.append(i)
        return non_extremes

    def set_indexes_under_limit(self, limit):
        dist = self.calculate_relative_distance()
        self.indexes_under_limit = np.where(dist < limit)[0]
        over = np.where(dist >= limit)[0]
        print("Removed images %s due to distance over limit: %s" % (np.add(1, over), limit))
        for i in over:
            print("%4s: distance: %.5f" % (i + 1, dist[i][0]))

    def get_indexes_under_limit(self):
        return self.indexes_under_limit


def nparray2str(data, precision=4, counter=float('Inf'), image=False):
    string = ""
    if image:
        formater = "%3.0f"
    else:
        formater = "%8." + str(precision) + "f\t"
    for j in data:
        counter -= 1
        for i in j:
            string += formater % i
        string += '\n'
        if counter < 0:
            break
    return string
