import os

import tensorflow as tf

from iProcessData import IProcessData, np2str, launch_tensorboard

import numpy as np

from scipy.stats import logistic

RUN_TYPE = "multi_layer_euler"
LIMIT_DATA = None

LOAD_MODEL = False
LOAD_MODEL_FILENAME = "single_layer_AUG0_STD0.5_E100000_B11_v0.ckpt"



TENSORBOARD_LOG = "D:/IProcess_Data/tensorflow_Logs"

"""#########################################################"""


def save_file_name(model_version=None):
    s = RUN_TYPE + "_"

    if LIMIT_DATA is not None:
        s += "limited" + str(LIMIT_DATA) + "_"

    s += "AUG{2}_STD{3}_E{0}_B{1}".format(EPOCHS, NUM_BATCH, AUG_COUNT, AUG_STD)

    if model_version is not None:
        s += "_v" + str(model_version)
    return s


def save_model():
    model_version = 0
    file_name = save_file_name(model_version) + ".ckpt"
    meta_files = [k for k in os.listdir("./tf_models") if ".meta" in k]
    while file_name + ".meta" in meta_files:
        model_version += 1
        file_name = save_file_name(model_version) + ".ckpt"
    saver = tf.train.Saver()
    saver.save(sess, "tf_models/" + file_name)
    print("*" * 100)
    print("Model saved in file: %s" % file_name)
    print("*" * 100)


def load_model(filename):
    saver = tf.train.Saver()
    model_file_name = "tf_models/" + filename
    saver.restore(sess, model_file_name)

    print("*" * 100)
    print("Model loaded from file: %s" % model_file_name)
    print("*" * 100)


def save_csv(prediction_data, default=False):

    if LOAD_MODEL:
        filename = LOAD_MODEL_FILENAME
        filename = filename[:-5] + ".csv"
    else:
        filename = save_file_name() + ".csv"

    if default:
        filename = "DEFAULT.csv"

    filepath = "tf_predictions/" + filename
    iprocess.save_prediction_to_file(prediction_data.T, filepath=filepath)
    print("*" * 100)
    print("CSV saved in file: %s" % filename)
    print("*" * 100)


def predict_accuracy(x_data, y_data):
    prediction_pred, prediction_loss = sess.run([dense2, loss], feed_dict={x: x_data, y_: y_data})
    print(sess.run([accuracy], feed_dict={x: x_data, y_: y_data}))
    print(np.mean(sess.run([accuracy], feed_dict={x: x_data, y_: y_data})))
    return prediction_pred



def run_log_folder(v):
    if LOAD_MODEL:
        return TENSORBOARD_LOG + "/" + LOAD_MODEL_FILENAME[:-5]
    else:
        return TENSORBOARD_LOG + "/" + save_file_name(v)

"""#################################################################"""
TEST_PERCENTAGE = 0.25

AUG_COUNT = 0
AUG_STD = 0

EPOCHS = 100
NUM_BATCH = int(525*(1-TEST_PERCENTAGE)*(1+AUG_COUNT)/25)


iprocess = IProcessData(coordinates=False)
iprocess.limit_examples(LIMIT_DATA)
iprocess.split_train_test(test_percentage=TEST_PERCENTAGE)
iprocess.augment_data(augmentation_count=AUG_COUNT, augmentation_std=AUG_STD)
iprocess.all_data_to_z_norm()

input_nodes = 12
output_nodes = 9
hidden_layer1 = 100

x = tf.placeholder(tf.float32, [None, input_nodes], name="X")

y_ = tf.placeholder(tf.float32, [None, output_nodes], name="Y_")


#dense1 = tf.contrib.layers.fully_connected(inputs=x, num_outputs=hidden_layer1, activation_fn=tf.nn.relu)
#dense2 = tf.contrib.layers.fully_connected(inputs=dense1, num_outputs=output_nodes)

dense1 = tf.layers.dense(inputs=x, units=hidden_layer1, activation=tf.nn.relu)
dense1_dropout = tf.layers.dropout(dense1, rate=0.2)
dense2 = tf.layers.dense(inputs=dense1_dropout, units=output_nodes)


# W1 = tf.Variable(tf.random_normal([input_nodes, hidden_layer1]), name="weight1")
# b1 = tf.Variable(tf.random_normal([hidden_layer1]), name="bias1")
# y1 = tf.nn.relu(tf.add(tf.matmul(x, W1), b1))
# #y1 = tf.matmul(x, W1)
#
# #y1_drop = tf.nn.dropout(y1, 0.5)
# y1_drop = y1
#
# W2 = tf.Variable(tf.random_normal([hidden_layer1, output_nodes]), name="weight2")
# b2 = tf.Variable(tf.random_normal([output_nodes]), name="bias2")
# y2 = tf.add(tf.matmul(y1_drop, W2), b2)
# #y2 = tf.matmul(y1, W2)


loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(logits=dense2, labels=y_))
accuracy = 1- tf.reduce_mean(tf.abs(tf.subtract(tf.sigmoid(dense2), y_)), axis=0)

#loss = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(logits=dense2, labels=y_))
#accuracy = tf.reduce_mean(tf.abs(tf.subtract(dense2, y_)), 0)

#loss = -tf.reduce_sum( (  (y_*tf.log(y2 + 1e-9)) + ((1-y_) * tf.log(1 - y2 + 1e-9)) )  , name='xentropy' )
# accuracy = tf.reduce_mean(tf.reduce_mean(tf.divide(tf.subtract(tf.sigmoid(y2), y_), tf.sigmoid(y2)), 0))

tf.summary.scalar("Loss", loss)
# tf.summary.scalar("Accuracy", accuracy)

train_step = tf.train.AdamOptimizer(0.0001).minimize(loss)
#train_step = tf.train.GradientDescentOptimizer(learning_rate=0.001).minimize(loss)
sess = tf.Session()
sess.run(tf.global_variables_initializer())


mergesum = tf.summary.merge_all()
writer = tf.summary.FileWriter(run_log_folder(1))
#launch_tensorboard(TENSORBOARD_LOG)
writer.add_graph(sess.graph)




if not LOAD_MODEL:

    # TRAINING
    batches = iprocess.get_training_batches(NUM_BATCH)
    print("Batch size:", batches['x'][0].shape[0], "\tnum bathes: ", len(batches['x']))
    counter = -1
    for epoch in range(EPOCHS):
        batches = iprocess.get_training_batches(NUM_BATCH)
        for batch in range(len(batches['x'])):
            _, batch_loss, summary = sess.run([train_step, loss, mergesum],
                                              feed_dict={x: batches['x'][batch], y_: batches['y'][batch]})

        writer.add_summary(summary, counter)
        counter += 1

        if epoch % (EPOCHS / 20) == 0:
            print('Epoch:\t', epoch)
            example_pred = predict_accuracy(iprocess.x_test, iprocess.y_test)

            print(50 * ' -')

    print("SAVING MODEL TO FILE")
    save_model()

else:
    load_model(LOAD_MODEL_FILENAME)

# PREDICTING
print(50 * ' *')
print("PREDICTING")
print("TRAINING DATA")
predict_accuracy(iprocess.x_train, iprocess.y_train)

print(50 * ' -')
print("TESTING DATA")
predict_accuracy(iprocess.x_test, iprocess.y_test)

print(50 * ' -')
print("COMPLETE DATASET")

dataset_prediction = sess.run([dense2], feed_dict={x: iprocess.data_to_z_norm(iprocess.dataset_X, input=True)})
dataset_prediction = logistic.cdf(dataset_prediction)
dataset_prediction = iprocess.data_from_z_norm(dataset_prediction, input=False)
save_csv(dataset_prediction)
save_csv(dataset_prediction, default=True)

iprocess.save_regression_results(save_file_name(), iprocess.y_test_unaugmented, (predict_accuracy(iprocess.x_test_unaugmented, iprocess.y_test_unaugmented)))


# TODO: SHOW PREDICTIONS FOR FIRST 10 TEST DATA
if True:
    print('\n' * 2)
    print(10 * '- predict ','\n')
    for a in range(5):
        get_example = iprocess.get_single_test(a)

        x_in = get_example['x']
        y_out = get_example['y']
        print("X\t", np2str(x_in))
        print("X\t", np2str(iprocess.data_from_z_norm(x_in, input=True)))
        print("Y_\t", np2str(y_out))
        print("Y_\t", np2str(iprocess.data_from_z_norm(y_out, input=False)))

        prediction = sess.run([dense2], feed_dict={x: [x_in]})
        #print("Yzs\t",prediction[0])
        #print("Yz\t",logistic.cdf(prediction))
        #print("Ys\t",np2str(iprocess.data_from_z_norm(prediction, input=False)[0][0]))
        print("Y\t",np2str(iprocess.data_from_z_norm(logistic.cdf(prediction[0][0]), input=False)))


        print(' -' * 50,'\n')
