import os

import tensorflow as tf

from iProcessData import IProcessData, np2str, launch_tensorboard

import numpy as np


RUN_TYPE = "multilayer_1"
NORMALIZE = True
LIMIT_DATA = None

LOAD_MODEL = False
LOAD_MODEL_FILENAME = "single_layer_norm_AUG500_STD50_E10000_B1000_v0.ckpt"

TEST_PERCENTAGE = 0.25

AUG_COUNT = 0
AUG_STD = 50

EPOCHS = 1000
NUM_BATCH = 10

TENSORBOARD_LOG = "D:/IProcess_Data/tensorflow_Logs"

"""#########################################################"""


def save_file_name(model_version=None):
    s = RUN_TYPE + "_"

    if NORMALIZE:
        s += "norm_"

    if LIMIT_DATA is not None:
        s += "limited" + str(LIMIT_DATA) + "_"

    s += "AUG{2}_STD{3}_E{0}_B{1}".format(EPOCHS, NUM_BATCH, AUG_COUNT, AUG_STD)

    if model_version is not None:
        s += "_v" + str(model_version)
    return s


def save_model():
    model_version = 0
    file_name = save_file_name(model_version) + ".ckpt"
    meta_files = [k for k in os.listdir("./tf_models") if ".meta" in k]
    while file_name + ".meta" in meta_files:
        model_version += 1
        file_name = save_file_name(model_version) + ".ckpt"
    saver = tf.train.Saver()
    saver.save(sess, "tf_models/" + file_name)
    print("*" * 100)
    print("Model saved in file: %s" % file_name)
    print("*" * 100)


def load_model(filename):
    saver = tf.train.Saver()
    model_file_name = "tf_models/" + filename
    saver.restore(sess, model_file_name)

    print("*" * 100)
    print("Model loaded from file: %s" % model_file_name)
    print("*" * 100)


def save_csv(prediction_data, default=False):

    if LOAD_MODEL:
        filename = LOAD_MODEL_FILENAME
        filename = filename[:-5] + ".csv"
    else:
        filename = save_file_name() + ".csv"

    if default:
        filename = "DEFAULT.csv"

    if NORMALIZE:
        prediction_data = iprocess.data_norm_to_output(prediction_data)

    filepath = "tf_predictions/" + filename
    iprocess.save_prediction_to_file(prediction_data.T, filepath=filepath)
    print("*" * 100)
    print("CSV saved in file: %s" % filename)
    print("*" * 100)


def predict_accuracy(x_data, y_data):
    prediction_pred, prediction_loss = sess.run([y, example_loss], feed_dict={x: x_data, y_: y_data})
    accuracy = np.mean(np.divide(np.subtract(prediction_pred, y_data), prediction_pred), 0)
    print("Loss:\t", round(prediction_loss / len(x_data), 4))
    iprocess.print_accuracy(accuracy)
    return prediction_pred

def run_log_folder(v):
    if LOAD_MODEL:
        return TENSORBOARD_LOG + "/" + LOAD_MODEL_FILENAME[:-5]
    else:
        return TENSORBOARD_LOG + "/" + save_file_name(v)

"""#################################################################"""



iprocess = IProcessData()
iprocess.limit_examples(LIMIT_DATA)
iprocess.split_train_test(test_percentage=TEST_PERCENTAGE)
iprocess.augment_data(augmentation_count=AUG_COUNT, augmentation_std=AUG_STD)
if NORMALIZE:
    iprocess.output_to_norm()


x = tf.placeholder(tf.float32, [None, 12], name="X")

y_ = tf.placeholder(tf.float32, [None, 15], name="Y_")

W1 = tf.Variable(tf.zeros([12, 50]), name="weight1")
b1 = tf.Variable(tf.zeros([50]), name="bias1")

W2 = tf.Variable(tf.zeros([50,15]), name="weight2")
b2 = tf.Variable(tf.zeros([15]), name="bias2")

y1 = tf.add(tf.matmul(x, W1), b1)
y = tf.add(tf.matmul(y1, W2), b2)



example_loss = tf.nn.l2_loss(tf.sub(y_, y))
# example_loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(y, y_))
tf.summary.scalar("loss", example_loss)

accuracy = tf.reduce_mean(tf.reduce_mean(tf.divide(tf.subtract(y, y_), y), 0))
tf.summary.scalar("Accuracy", accuracy)

train_step = tf.train.AdamOptimizer(0.00000001, epsilon=1e-5).minimize(example_loss)
sess = tf.Session()
sess.run(tf.global_variables_initializer())


mergesum = tf.summary.merge_all()
writer = tf.summary.FileWriter(run_log_folder(1))
#launch_tensorboard(TENSORBOARD_LOG)
writer.add_graph(sess.graph)




if not LOAD_MODEL:

    # TRAINING
    batches = iprocess.get_training_batches(NUM_BATCH)
    print("Batch size:", batches['x'][0].shape[0], "\tnum bathes: ", len(batches['x']))
    counter = 0
    for epoch in range(EPOCHS):
        for batch in range(len(batches['x'])):
            _, batch_loss, summary = sess.run([train_step, example_loss, mergesum],
                                     feed_dict={x: batches['x'][batch], y_: batches['y'][batch]})

            writer.add_summary(summary, counter)
            counter += 1

        if epoch % (EPOCHS / 20) == 0:
            print('Epoch:\t', epoch)

            predict_accuracy(iprocess.x_test, iprocess.y_test)
            print(sess.run([accuracy], feed_dict={x: iprocess.x_test, y_: iprocess.y_test}))

            print(50 * ' -')

    print("SAVING MODEL TO FILE")
    save_model()

else:
    load_model(LOAD_MODEL_FILENAME)

# PREDICTING
print(50 * ' *')
print("PREDICTING")
print("TRAINING DATA")
predict_accuracy(iprocess.x_train, iprocess.y_train)

print(50 * ' -')
print("TESTING DATA")
predict_accuracy(iprocess.x_test, iprocess.y_test)

print(50 * ' -')
print("COMPLETE DATASET")
dataset_prediction = predict_accuracy(iprocess.dataset_X, iprocess.dataset_y)
save_csv(dataset_prediction)
save_csv(dataset_prediction, default=True)

iprocess.save_regression_results(save_file_name(), iprocess.y_test_unaugmented, predict_accuracy(iprocess.x_test_unaugmented, iprocess.y_test_unaugmented))


# TODO: SHOW PREDICTIONS FOR FIRST 10 TEST DATA
if False:
    print('\n' * 2)
    print(10 * '- predict ')
    for a in range(10):
        get_example = iprocess.get_single_test(a)

        x_in = get_example['x']
        y_out = get_example['y']
        print("\nX\t", np2str(x_in))
        print("Y_\t", np2str(y_out))

        example_pred = predict_accuracy([x_in], [y_out])

        if (NORMALIZE):
            print("Converted")
            print("Y_\t", np2str(iprocess.data_norm_to_output(y_out)))
            print("y\t", np2str(iprocess.data_norm_to_output(example_pred[0])))
            np2str(y_out)

        print(' -' * 50)
