import numpy as np
import tensorflow as tf
from scipy import misc
from scipy.special import expit

from CNN.csvData import CSVData
from CNN.data_loader import denorm, cos_sin2deg, nparray2str

PIXELS = 227
CHANNELS = 4

IMAGE_FOLDER = "cnn_prediction_live/"
RELATIVE_FILE = "relative_data.csv"
PREDICTION_FILE = "prediction.txt"
COLOR = "Color.png"
DEPTH = "Depth.png"

TENSORBOARD_MODELS = "D:/IProcess_Data/cnn_models/"
model = "model_1497508826.ckpt-15"


def png_to_nparray_x(image_d=None, image_c=None):
    png_d = None
    png_c = None

    if image_d is not None:
        png_d = misc.imread(image_d)
        png_d = misc.imresize(png_d, (PIXELS, PIXELS))
        if image_c is None:
            return png_d

    if image_c is not None:
        png_c = misc.imread(image_c)
        png_c = misc.imresize(png_c, (PIXELS, PIXELS, 3))
        if image_d is None:
            return png_c

    return np.dstack((png_c, png_d))


def denomralize_outputs(output):
    output_copy = np.copy(output)
    output_copy[:, :3] = denorm(output_copy[:, :3], -250, 250)
    if output_copy.shape[1] > 3:
        output_copy[:, 3:] = denorm(output_copy[:, 3:], -1, 1)
        output_copy = cos_sin2deg(output_copy)
    return output_copy


def load_model(sess, filename):
    model_file_name = TENSORBOARD_MODELS + filename
    saver = tf.train.import_meta_graph(model_file_name + ".meta")
    saver.restore(sess, model_file_name)

    print("*" * 100)
    print("Model loaded from file: %s" % model_file_name)
    print("*" * 100)


# Start TF session
sess = tf.Session()

# Load image from folder and set to correct dimensions
print("LOADING IMAGE FROM FOLDER: ", IMAGE_FOLDER)
png_array = png_to_nparray_x(image_d=IMAGE_FOLDER + DEPTH, image_c=IMAGE_FOLDER + COLOR)
print("IMAGE LOADED WITH SHAPE:", png_array.shape)

# Load relative data from csv file
print("RELATIVE DATA LOADED FROM: ", RELATIVE_FILE)
csv = CSVData(IMAGE_FOLDER + RELATIVE_FILE, densodata=False)
predicted_6dof = np.append(csv.coordinates, [[0, 0, 0]], 1)

# Load model and set variables
load_model(sess, model)
for i in tf.trainable_variables(): print(i)
#    for i in tf.get_default_graph().get_all_collection_keys(): print(i)

net = tf.get_collection("net")[0]
x = tf.get_collection("X")[0]
y = tf.get_collection("Y_")[0]
abs_distance = tf.get_collection("abs_distance")[0]
print("\nMODEL LOADED AND TF-VARIABLES ASSIGNED")

# reshape due to single image input
png_array = np.reshape(png_array, [-1, PIXELS, PIXELS, CHANNELS])
# run image through network
prediction = sess.run(net, feed_dict={x: png_array})
print("\nPREDICTION COMPLETED USING MODEL")
# convert from output to 6DOF
prediction = expit(prediction)
prediction = denomralize_outputs(prediction)

print("\nOUTPUT FROM MODEL AS RELATIVE 6DOF")
print(nparray2str(prediction))

# convert from relative to denso 6DOF by adding prediction to center of image data
predicted_6dof = np.add(predicted_6dof, prediction)

print("\nDENSO 6DOF")
print(nparray2str(predicted_6dof))

with open(IMAGE_FOLDER + PREDICTION_FILE, "w") as f:
    string = ""
    for r in predicted_6dof[0]:
        string += str(r) + "\t"
    f.write(string)
print("\nDENSO 6DOF SAVED TO FILE: ", PREDICTION_FILE)
